import json
import platform
from os import getenv, mkdir, path
from urllib.parse import quote_plus, unquote_plus

import aiofiles
import aiofiles.os
import keyring
from cryptography.fernet import Fernet

SYSTEM = platform.system()
if SYSTEM == "Linux":
    PATHWAY = f"{getenv('HOME')}/.config/omitme"
elif SYSTEM == "Windows":
    PATHWAY = f"{getenv('APPDATA')}\\omitme"
elif SYSTEM == "Darwin":
    PATHWAY = f"/Users/{getenv('HOME')}/Library/Application Support/omitme"
else:
    raise Exception("Platform not supported.")


__key_raw = keyring.get_password("omitme", "fernet")
if __key_raw is None:
    RAW_FERNET_KEY = Fernet.generate_key()
    keyring.set_password("omitme", "fernet", RAW_FERNET_KEY.hex())
else:
    RAW_FERNET_KEY = bytes.fromhex(__key_raw)

if not path.exists(PATHWAY):
    mkdir(PATHWAY)


class Accounts:
    def __init__(self, platform: str) -> None:
        self.__platform = platform

    @property
    def __safe_platform(self) -> str:
        return quote_plus(self.__platform)

    def __json_store_path(self, username: str) -> str:
        return path.join(PATHWAY, f"{self.__safe_platform}-{quote_plus(username)}.bin")

    async def remove(self, username: str) -> None:
        await aiofiles.os.remove(self.__json_store_path(username))

    async def add(self, username: str, session: dict) -> None:
        async with aiofiles.open(
            self.__json_store_path(username),
            "wb+",
        ) as f_:
            await f_.write(Fernet(RAW_FERNET_KEY).encrypt(json.dumps(session).encode()))

    async def get(self, username: str) -> dict:
        async with aiofiles.open(self.__json_store_path(username), "rb") as f_:
            return json.loads(Fernet(RAW_FERNET_KEY).decrypt(await f_.read()))

    async def list(self) -> list[str]:
        return [
            unquote_plus(file.removeprefix(f"{self.__platform}-").removesuffix(".bin"))
            for file in await aiofiles.os.listdir(PATHWAY)
            if file.startswith(self.__platform)
        ]



import asyncio
from typing import Optional

import click

from omitme.platforms import PLATFORMS
from omitme.util.targets import SelectionInput


@click.group()
def main() -> None:
    """Interact with OmitMe from your terminal!"""

    pass


loop = asyncio.get_event_loop()


for platform in PLATFORMS:

    @main.group(platform.alias, help=platform.description)
    def platform_group() -> None:
        pass

    @platform_group.command("login", help="Log into a new account")
    def login(platform=platform) -> None:
        loop.run_until_complete(platform().handle_login())  # type: ignore

    @platform_group.command("accounts", help="List accounts already logged in")
    def accounts(platform=platform) -> None:
        accounts = loop.run_until_complete(platform().list_accounts())

        for account in accounts:
            click.echo(f"- {account}")

    @platform_group.command("logout", help="Logout the given account")
    @click.option(
        "--account",
        prompt="The account you wish the remove",
        help="Provide the account ID you want to remove",
        required=True,
    )
    def logout(account: str, platform=platform) -> None:
        loop.run_until_complete(platform().remove_account(account))

    @platform_group.group("target", help="Run specific target on the selected platform")
    def target() -> None:
        pass

    for method, meta in platform._target_methods:

        @target.command(meta.action.replace(" ", "-"), help=meta.description)
        @click.option(
            "--account",
            prompt="The account to use with the target",
            help="Provide the account you want to commit a action on",
            required=True,
        )
        @click.option(
            "--echo",
            default=True,
            type=click.BOOL,
            help="Should logs be echo into the terminal in real-time",
        )
        def handle_target(
            account: str, echo: bool, platform=platform, meta=meta, method=method
        ) -> None:
            platform_init = platform()
            loop.run_until_complete(platform_init.load_account(account.strip()))

            parameters = {}

            if meta.inputs:
                for input_ in meta.inputs:
                    if isinstance(input_, SelectionInput):
                        selections = loop.run_until_complete(
                            input_.run(platform_init, platform_init._session)
                        )

                        selection_choices = []
                        indexed_selection = {}
                        for selection in selections:
                            display_format = f"{input_.format}".format_map(selection)

                            selection_choices.append(display_format)
                            indexed_selection[display_format] = selection

                        click.echo(", ".join(selection_choices))

                        chosen = click.prompt(
                            f"Select {input_.name} (comma separated)",
                        )

                        parameters[input_.parameter] = []
                        for selected in chosen.split(","):
                            selected = selected.strip()

                            if selected not in indexed_selection:
                                raise Exception(
                                    f"{selected} not found in {input_.name}"
                                )

                            parameters[input_.parameter].append(
                                indexed_selection[selected]
                            )

            async def handle_events(platform_init, method, parameters) -> None:
                async for event in getattr(platform_init, method.__name__)(
                    **parameters
                ):
                    if echo:
                        click.echo(event)

            loop.run_until_complete(handle_events(platform_init, method, parameters))



from typing import AsyncIterator

import httpx
from selenium.common.exceptions import TimeoutException
from seleniumwire import webdriver
from seleniumwire.request import HTTPHeaders

from omitme.errors import LoginError
from omitme.util.accounts import Accounts
from omitme.util.events import CheckingEvent, CompletedEvent, FailEvent, OmittedEvent
from omitme.util.platform import Platform
from omitme.util.targets import login, target
from omitme.util.wait_for import wait_until_or_close


class Reddit(Platform):
    login_url = (
        "https://www.reddit.com/login/?dest=https%3A%2F%2Fwww.reddit.com%2Fsettings"
    )
    api_url = "https://oauth.reddit.com"
    alias = "reddit"
    description = "Manage your reddit data"
    icon = "reddit.png"

    @login
    async def handle_login(
        self, driver: webdriver.Chrome, accounts: Accounts
    ) -> httpx.AsyncClient:
        driver.get(self.login_url)

        def get_headers_with_token(current: webdriver.Chrome) -> dict | bool:
            for request in current.requests:
                if "Authorization" in request.headers:
                    headers = {}
                    if request.response:
                        headers |= request.response.headers

                    headers |= request.headers

                    return headers
            return False

        try:
            headers: dict = wait_until_or_close(driver, get_headers_with_token)
        except TimeoutException:
            raise LoginError()

        headers.pop("content-length")
        headers.pop("content-type")

        to_save = {
            "headers": headers,
            "params": {
                "redditWebClient": "desktop2x",
                "app": "desktop2x-client-production",
            },
        }

        session = httpx.AsyncClient(**to_save, base_url=self.api_url)  # type: ignore

        await accounts.add(await self._get_username(session), to_save)

        return session

    async def _delete_post(
        self, session: httpx.AsyncClient, type_id: str, post_id: str
    ) -> None:
        await session.post(
            "/api/del",
            params={"raw_json": "1", "gilding_detail": "1"},
            data={"id": f"{type_id}_{post_id}"},
        )

    async def _get_username(self, session: httpx.AsyncClient) -> str:
        resp = await session.get(
            "https://gateway.reddit.com/desktopapi/v1/prefs",
            params={"include": "identity"},
        )

        return resp.json()["account"]["displayText"]

    @target("posts delete all", description="Delete all reddit posts")
    async def handle_delete_posts(
        self, session: httpx.AsyncClient
    ) -> AsyncIterator[OmittedEvent | CheckingEvent | FailEvent | CompletedEvent]:
        resp = await session.get(
            f"/user/{await self._get_username(session)}/submitted",
            params={"timeframe": "all", "limit": "100"},
        )

        for post in resp.json()["data"]["children"]:
            await self._delete_post(session, post["kind"], post["data"]["id"])

            yield OmittedEvent(
                channel=post["data"]["subreddit"], content=post["data"]["selftext"]
            )

        yield CompletedEvent()



from typing import Any, Callable

from selenium.common.exceptions import TimeoutException, WebDriverException
from selenium.webdriver.support.wait import WebDriverWait
from seleniumwire import webdriver

from omitme.errors import LoginError


def wait_until_or_close(
    driver: webdriver.Chrome,
    condition: Callable,
    timeout: int = 920,
) -> Any:
    def close_condition(driver: webdriver.Chrome) -> bool:
        try:
            driver.title
            return False
        except WebDriverException:
            raise LoginError()

    try:
        return WebDriverWait(driver, timeout).until(
            lambda driver: condition(driver) or close_condition(driver)
        )
    except TimeoutException:
        pass



from functools import wraps
from typing import TYPE_CHECKING, Any, AsyncIterator, Callable, Coroutine, List, cast

import httpx
from pydantic import BaseModel
from seleniumwire import webdriver

from omitme.errors import LoginRequiredError

if TYPE_CHECKING:
    from omitme.util.platform import Platform


class SelectionInput(BaseModel):
    parameter: str
    run: Callable[[Any, httpx.AsyncClient], Coroutine[Any, Any, Any]]
    required: bool = False
    format: str
    name: str


class Target(BaseModel):
    description: str | None = None
    action: str
    inputs: List[SelectionInput] | None = None


async def raise_on_4xx_5xx(response: httpx.Response):
    response.raise_for_status()


def login(func: Callable) -> Callable:
    @wraps(func)
    async def _implement(*args, **kwargs) -> Any:
        self_ = cast("Platform", args[0])

        driver = webdriver.Chrome(options=self_.webdriver_options)

        self_._session = cast(
            httpx.AsyncClient,
            await func(self_, driver=driver, accounts=self_._account),
        )
        self_._session.base_url = self_.api_url
        self_._session.event_hooks = {"response": [raise_on_4xx_5xx]}
        self_._session.headers["User-Agent"] = self_.user_agent

        driver.quit()

        return self_._session

    return _implement


def target(
    action: str,
    description: str | None = None,
    inputs: List[SelectionInput] | None = None,
) -> Callable:
    def _func(func: Callable) -> Callable:
        func._target_data = Target(
            description=description, action=action, inputs=inputs
        )

        @wraps(func)
        async def _implement(*args, **kwargs) -> AsyncIterator[Any]:
            self_ = cast("Platform", args[0])

            if not self_._session:
                raise LoginRequiredError()

            async for result in func(self_, session=self_._session, **kwargs):
                yield result

        return _implement

    return _func
